@@ -1,38 +1,41 @@
-# Copyright 2012-2019, Andrey Kislyuk and argcomplete contributors.
+# Copyright 2012-2021, Andrey Kislyuk and argcomplete contributors.
 # Licensed under the Apache License. See https://github.com/kislyuk/argcomplete for more info.
 
-from __future__ import absolute_import, division, print_function, unicode_literals
+import argparse
+import contextlib
+import os
+import sys
+from typing import Dict, List, Tuple
 
-import os, sys, argparse, contextlib
-from . import completers, my_shlex as shlex
-from .compat import USING_PYTHON2, str, sys_encoding, ensure_str, ensure_bytes
+from . import completers
+from . import my_shlex as shlex
 from .completers import FilesCompleter, SuppressCompleter
-from .my_argparse import IntrospectiveArgumentParser, action_is_satisfied, action_is_open, action_is_greedy
-from .shellintegration import shellcode
+from .my_argparse import IntrospectiveArgumentParser, action_is_greedy, action_is_open, action_is_satisfied
+from .shell_integration import shellcode  # noqa
 
 _DEBUG = "_ARC_DEBUG" in os.environ
 
 debug_stream = sys.stderr
 
+
 def debug(*args):
     if _DEBUG:
-        if USING_PYTHON2:
-            # debug_stream has to be binary mode in Python 2.
-            # Attempting to write unicode directly uses the default ascii conversion.
-            # Convert any unicode to bytes, leaving non-string input alone.
-            args = [ensure_bytes(x) if isinstance(x, str) else x for x in args]
         print(file=debug_stream, *args)
 
+
 BASH_FILE_COMPLETION_FALLBACK = 79
 BASH_DIR_COMPLETION_FALLBACK = 80
 
-safe_actions = (argparse._StoreAction,
-                argparse._StoreConstAction,
-                argparse._StoreTrueAction,
-                argparse._StoreFalseAction,
-                argparse._AppendAction,
-                argparse._AppendConstAction,
-                argparse._CountAction)
+safe_actions = (
+    argparse._StoreAction,
+    argparse._StoreConstAction,
+    argparse._StoreTrueAction,
+    argparse._StoreFalseAction,
+    argparse._AppendAction,
+    argparse._AppendConstAction,
+    argparse._CountAction,
+)
+
 
 @contextlib.contextmanager
 def mute_stdout():
@@ -43,6 +46,7 @@ def mute_stdout():
     finally:
         sys.stdout = stdout
 
+
 @contextlib.contextmanager
 def mute_stderr():
     stderr = sys.stderr
@@ -53,12 +57,15 @@ def mute_stderr():
         sys.stderr.close()
         sys.stderr = stderr
 
+
 class ArgcompleteException(Exception):
     pass
 
+
 def split_line(line, point=None):
     if point is None:
         point = len(line)
+    line = line[:point]
     lexer = shlex.shlex(line, posix=True)
     lexer.whitespace_split = True
     lexer.wordbreaks = os.environ.get("_ARGCOMPLETE_COMP_WORDBREAKS", "")
@@ -100,21 +107,34 @@ def split_line(line, point=None):
             if lexer.instream.tell() >= point:
                 return split_word(lexer.token)
             else:
-                msg = ("Unexpected internal state. "
-                       "Please report this bug at https://github.com/kislyuk/argcomplete/issues.")
+                msg = (
+                    "Unexpected internal state. "
+                    "Please report this bug at https://github.com/kislyuk/argcomplete/issues."
+                )
                 raise ArgcompleteException(msg)
 
+
 def default_validator(completion, prefix):
     return completion.startswith(prefix)
 
+
 class CompletionFinder(object):
     """
     Inherit from this class if you wish to override any of the stages below. Otherwise, use
     ``argcomplete.autocomplete()`` directly (it's a convenience instance of this class). It has the same signature as
     :meth:`CompletionFinder.__call__()`.
     """
-    def __init__(self, argument_parser=None, always_complete_options=True, exclude=None, validator=None,
-                 print_suppressed=False, default_completer=FilesCompleter(), append_space=None):
+
+    def __init__(
+        self,
+        argument_parser=None,
+        always_complete_options=True,
+        exclude=None,
+        validator=None,
+        print_suppressed=False,
+        default_completer=FilesCompleter(),
+        append_space=None,
+    ):
         self._parser = argument_parser
         self.always_complete_options = always_complete_options
         self.exclude = exclude
@@ -123,15 +143,24 @@ class CompletionFinder(object):
         self.validator = validator
         self.print_suppressed = print_suppressed
         self.completing = False
-        self._display_completions = {}
+        self._display_completions: Dict[Tuple[str], str] = {}
         self.default_completer = default_completer
         if append_space is None:
             append_space = os.environ.get("_ARGCOMPLETE_SUPPRESS_SPACE") != "1"
         self.append_space = append_space
 
-    def __call__(self, argument_parser, always_complete_options=True, exit_method=os._exit, output_stream=None,
-                 exclude=None, validator=None, print_suppressed=False, append_space=None,
-                 default_completer=FilesCompleter()):
+    def __call__(
+        self,
+        argument_parser,
+        always_complete_options=True,
+        exit_method=os._exit,
+        output_stream=None,
+        exclude=None,
+        validator=None,
+        print_suppressed=False,
+        append_space=None,
+        default_completer=FilesCompleter(),
+    ):
         """
         :param argument_parser: The argument parser to autocomplete on
         :type argument_parser: :class:`argparse.ArgumentParser`
@@ -169,9 +198,15 @@ class CompletionFinder(object):
         added to argcomplete.safe_actions, if their values are wanted in the ``parsed_args`` completer argument, or
         their execution is otherwise desirable.
         """
-        self.__init__(argument_parser, always_complete_options=always_complete_options, exclude=exclude,
-                      validator=validator, print_suppressed=print_suppressed, append_space=append_space,
-                      default_completer=default_completer)
+        self.__init__(  # type: ignore
+            argument_parser,
+            always_complete_options=always_complete_options,
+            exclude=exclude,
+            validator=validator,
+            print_suppressed=print_suppressed,
+            append_space=append_space,
+            default_completer=default_completer,
+        )
 
         if "_ARGCOMPLETE" not in os.environ:
             # not an argument completion invocation
@@ -180,13 +215,20 @@ class CompletionFinder(object):
         global debug_stream
         try:
             debug_stream = os.fdopen(9, "w")
-        except:
+        except Exception:
             debug_stream = sys.stderr
+        debug()
+
+        if output_stream is None:
+            filename = os.environ.get("_ARGCOMPLETE_STDOUT_FILENAME")
+            if filename is not None:
+                debug("Using output file {}".format(filename))
+                output_stream = open(filename, "w")
 
         if output_stream is None:
             try:
-                output_stream = os.fdopen(8, "wb")
-            except:
+                output_stream = os.fdopen(8, "w")
+            except Exception:
                 debug("Unable to open fd 8 for writing, quitting")
                 exit_method(1)
 
@@ -199,10 +241,14 @@ class CompletionFinder(object):
             debug("Invalid value for IFS, quitting [{v}]".format(v=ifs))
             exit_method(1)
 
+        dfs = os.environ.get("_ARGCOMPLETE_DFS")
+        if dfs and len(dfs) != 1:
+            debug("Invalid value for DFS, quitting [{v}]".format(v=dfs))
+            exit_method(1)
+
         comp_line = os.environ["COMP_LINE"]
         comp_point = int(os.environ["COMP_POINT"])
 
-        comp_line = ensure_str(comp_line)
         cword_prequote, cword_prefix, cword_suffix, comp_words, last_wordbreak_pos = split_line(comp_line, comp_point)
 
         # _ARGCOMPLETE is set by the shell script to tell us where comp_words
@@ -217,17 +263,28 @@ class CompletionFinder(object):
             # Special case for when the current word is "--optional=PARTIAL_VALUE". Give the optional to the parser.
             comp_words.append(cword_prefix.split("=", 1)[0])
 
-        debug("\nLINE: {!r}".format(comp_line),
-              "\nPOINT: {!r}".format(comp_point),
-              "\nPREQUOTE: {!r}".format(cword_prequote),
-              "\nPREFIX: {!r}".format(cword_prefix),
-              "\nSUFFIX: {!r}".format(cword_suffix),
-              "\nWORDS:", comp_words)
+        debug(
+            "\nLINE: {!r}".format(comp_line),
+            "\nPOINT: {!r}".format(comp_point),
+            "\nPREQUOTE: {!r}".format(cword_prequote),
+            "\nPREFIX: {!r}".format(cword_prefix),
+            "\nSUFFIX: {!r}".format(cword_suffix),
+            "\nWORDS:",
+            comp_words,
+        )
 
         completions = self._get_completions(comp_words, cword_prefix, cword_prequote, last_wordbreak_pos)
 
+        if dfs:
+            display_completions = {
+                key_part: value.replace(ifs, " ") if value else ""
+                for key, value in self._display_completions.items()
+                for key_part in key
+            }
+            completions = [dfs.join((key, display_completions.get(key) or "")) for key in completions]
+
         debug("\nReturning completions:", completions)
-        output_stream.write(ifs.join(completions).encode(sys_encoding))
+        output_stream.write(ifs.join(completions))
         output_stream.flush()
         debug_stream.flush()
         exit_method(0)
@@ -238,10 +295,6 @@ class CompletionFinder(object):
         parsed_args = argparse.Namespace()
         self.completing = True
 
-        if USING_PYTHON2:
-            # Python 2 argparse only properly works with byte strings.
-            comp_words = [ensure_bytes(word) for word in comp_words]
-
         try:
             debug("invoking parser with", comp_words[1:])
             with mute_stderr():
@@ -267,8 +320,8 @@ class CompletionFinder(object):
         figure out which subparsers need to be activated (then recursively monkey-patch those).
         We save all active ArgumentParsers to extract all their possible option names later.
         """
-        self.active_parsers = []
-        self.visited_positionals = []
+        self.active_parsers: List[argparse.ArgumentParser] = []
+        self.visited_positionals: List[argparse.Action] = []
 
         completer = self
 
@@ -280,17 +333,15 @@ class CompletionFinder(object):
                 return
 
             classname = "MonkeyPatchedIntrospectiveArgumentParser"
-            if USING_PYTHON2:
-                classname = bytes(classname)
+
             parser.__class__ = type(classname, (IntrospectiveArgumentParser, parser.__class__), {})
 
             for action in parser._actions:
-
                 if hasattr(action, "_orig_class"):
                     continue
 
                 # TODO: accomplish this with super
-                class IntrospectAction(action.__class__):
+                class IntrospectAction(action.__class__):  # type: ignore
                     def __call__(self, parser, namespace, values, option_string=None):
                         debug("Action stub called on", self)
                         debug("\targs:", parser, namespace, values, option_string)
@@ -321,16 +372,16 @@ class CompletionFinder(object):
         return self.active_parsers
 
     def _get_subparser_completions(self, parser, cword_prefix):
-        def filter_aliases(metavar, dest, prefix):
-            if not metavar:
-                return dest if dest and dest.startswith(prefix) else ""
+        def filter_aliases(aliases, prefix):
+            return tuple(x for x in aliases if x.startswith(prefix))
 
-            # metavar combines dest and aliases with ",".
-            a = metavar.replace(",", "").split()
-            return " ".join(x for x in a if x.startswith(prefix))
+        aliases_by_parser: Dict[argparse.ArgumentParser, List[str]] = {}
+        for key in parser.choices.keys():
+            p = parser.choices[key]
+            aliases_by_parser.setdefault(p, []).append(key)
 
         for action in parser._get_subactions():
-            subcmd_with_aliases = filter_aliases(action.metavar, action.dest, cword_prefix)
+            subcmd_with_aliases = filter_aliases(aliases_by_parser[parser.choices[action.dest]], cword_prefix)
             if subcmd_with_aliases:
                 self._display_completions[subcmd_with_aliases] = action.help
 
@@ -339,9 +390,9 @@ class CompletionFinder(object):
 
     def _include_options(self, action, cword_prefix):
         if len(cword_prefix) > 0 or self.always_complete_options is True:
-            return [ensure_str(opt) for opt in action.option_strings if ensure_str(opt).startswith(cword_prefix)]
-        long_opts = [ensure_str(opt) for opt in action.option_strings if len(opt) > 2]
-        short_opts = [ensure_str(opt) for opt in action.option_strings if len(opt) <= 2]
+            return [opt for opt in action.option_strings if opt.startswith(cword_prefix)]
+        long_opts = [opt for opt in action.option_strings if len(opt) > 2]
+        short_opts = [opt for opt in action.option_strings if len(opt) <= 2]
         if self.always_complete_options == "long":
             return long_opts if long_opts else short_opts
         elif self.always_complete_options == "short":
@@ -349,11 +400,10 @@ class CompletionFinder(object):
         return []
 
     def _get_option_completions(self, parser, cword_prefix):
-        self._display_completions.update(
-            [[" ".join(ensure_str(x) for x in action.option_strings
-                       if ensure_str(x).startswith(cword_prefix)), action.help]
-             for action in parser._actions
-             if action.option_strings])
+        for action in parser._actions:
+            if action.option_strings:
+                active_option_strings = tuple(str(x) for x in action.option_strings if x.startswith(cword_prefix))
+                self._display_completions[active_option_strings] = action.help  # type: ignore
 
         option_completions = []
         for action in parser._actions:
@@ -428,27 +478,33 @@ class CompletionFinder(object):
                     completer = self.default_completer
 
             if completer:
+                if isinstance(completer, SuppressCompleter) and completer.suppress():
+                    continue
+
                 if callable(completer):
-                    completions_from_callable = [c for c in completer(
-                        prefix=cword_prefix, action=active_action, parser=parser, parsed_args=parsed_args)
-                        if self.validator(c, cword_prefix)]
+                    completions_from_callable = [
+                        c
+                        for c in completer(
+                            prefix=cword_prefix, action=active_action, parser=parser, parsed_args=parsed_args
+                        )
+                        if self.validator(c, cword_prefix)
+                    ]
 
                     if completions_from_callable:
                         completions += completions_from_callable
-                        if isinstance(completer, completers.ChoicesCompleter):
-                            self._display_completions.update(
-                                [[x, active_action.help] for x in completions_from_callable])
-                        else:
-                            self._display_completions.update(
-                                [[x, ""] for x in completions_from_callable])
+                        for x in completions_from_callable:
+                            if isinstance(completer, completers.ChoicesCompleter):
+                                self._display_completions[(x,)] = active_action.help
+                            else:
+                                self._display_completions[(x,)] = ""
                 else:
                     debug("Completer is not callable, trying the readline completer protocol instead")
                     for i in range(9999):
-                        next_completion = completer.complete(cword_prefix, i)
+                        next_completion = completer.complete(cword_prefix, i)  # type: ignore
                         if next_completion is None:
                             break
                         if self.validator(next_completion, cword_prefix):
-                            self._display_completions.update({next_completion: ""})
+                            self._display_completions.update({(next_completion,): ""})  # type: ignore
                             completions.append(next_completion)
                 if optional_prefix:
                     completions = [optional_prefix + "=" + completion for completion in completions]
@@ -477,8 +533,9 @@ class CompletionFinder(object):
         if isinstance(next_positional, argparse._SubParsersAction):
             completions += self._get_subparser_completions(next_positional, cword_prefix)
 
-        completions = self._complete_active_option(active_parser, next_positional, cword_prefix, parsed_args,
-                                                   completions)
+        completions = self._complete_active_option(
+            active_parser, next_positional, cword_prefix, parsed_args, completions
+        )
         debug("active options:", completions)
         debug("display completions:", self._display_completions)
 
@@ -510,21 +567,19 @@ class CompletionFinder(object):
 
     def filter_completions(self, completions):
         """
-        Ensures collected completions are Unicode text, de-duplicates them, and excludes those specified by ``exclude``.
-        Returns the filtered completions as an iterable.
+        De-duplicates completions and excludes those specified by ``exclude``.
+        Returns the filtered completions as a list.
 
         This method is exposed for overriding in subclasses; there is no need to use it directly.
         """
-        # On Python 2, we have to make sure all completions are unicode objects before we continue and output them.
-        # Otherwise, because python disobeys the system locale encoding and uses ascii as the default encoding, it will
-        # try to implicitly decode string objects using ascii, and fail.
-        completions = [ensure_str(c) for c in completions]
-
-        # De-duplicate completions and remove excluded ones
-        if self.exclude is None:
-            self.exclude = set()
-        seen = set(self.exclude)
-        return [c for c in completions if c not in seen and not seen.add(c)]
+        filtered_completions = []
+        for completion in completions:
+            if self.exclude is not None:
+                if completion in self.exclude:
+                    continue
+            if completion not in filtered_completions:
+                filtered_completions.append(completion)
+        return filtered_completions
 
     def quote_completions(self, completions, cword_prequote, last_wordbreak_pos):
         """
@@ -546,7 +601,7 @@ class CompletionFinder(object):
             # This workaround has the same effect as __ltrim_colon_completions in bash_completion
             # (extended to characters other than the colon).
             if last_wordbreak_pos:
-                completions = [c[last_wordbreak_pos + 1:] for c in completions]
+                completions = [c[last_wordbreak_pos + 1 :] for c in completions]
             special_chars += "();<>|&!`$* \t\n\"'"
         elif cword_prequote == '"':
             special_chars += '"`$!'
@@ -577,7 +632,7 @@ class CompletionFinder(object):
     def rl_complete(self, text, state):
         """
         Alternate entry point for using the argcomplete completer in a readline-based REPL. See also
-        `rlcompleter <https://docs.python.org/2/library/rlcompleter.html#completer-objects>`_.
+        `rlcompleter <https://docs.python.org/3/library/rlcompleter.html#completer-objects>`_.
         Usage:
 
         .. code-block:: python
@@ -590,14 +645,12 @@ class CompletionFinder(object):
             readline.set_completer(completer.rl_complete)
             readline.parse_and_bind("tab: complete")
             result = input("prompt> ")
-
-        (Use ``raw_input`` instead of ``input`` on Python 2, or use `eight <https://github.com/kislyuk/eight>`_).
         """
         if state == 0:
             cword_prequote, cword_prefix, cword_suffix, comp_words, first_colon_pos = split_line(text)
             comp_words.insert(0, sys.argv[0])
             matches = self._get_completions(comp_words, cword_prefix, cword_prequote, first_colon_pos)
-            self._rl_matches = [text + match[len(cword_prefix):] for match in matches]
+            self._rl_matches = [text + match[len(cword_prefix) :] for match in matches]
 
         if state < len(self._rl_matches):
             return self._rl_matches[state]
@@ -635,7 +688,8 @@ class CompletionFinder(object):
             readline.set_completion_display_matches_hook(display_completions)
 
         """
-        return self._display_completions
+        return {" ".join(k): v for k, v in self._display_completions.items()}
+
 
 class ExclusiveCompletionFinder(CompletionFinder):
     @staticmethod
@@ -652,12 +706,17 @@ class ExclusiveCompletionFinder(CompletionFinder):
 
         return False
 
+
 autocomplete = CompletionFinder()
 autocomplete.__doc__ = """ Use this to access argcomplete. See :meth:`argcomplete.CompletionFinder.__call__()`. """
 
+
 def warn(*args):
     """
     Prints **args** to standard error when running completions. This will interrupt the user's command line interaction;
     use it to indicate an error condition that is preventing your completer from working.
     """
-    print("\n", file=debug_stream, *args)
+    # Don't be tempted to use `print("\n",..., *args)`,
+    # as that will indent **args** by one space character
+    print(file=debug_stream)
+    print(file=debug_stream, *args)
